# Copyright (c) 2015, Workiva Inc.  All rights reserved
# Copyright (c) 2015, XBRL US Inc.  All rights reserved
import json
import os.path

_LOADED_MESSAGES = None


def _load_messages(filename=None):
    """
    Loads the messages from the filename passed in.  If nothing is passed in,
    it defaults to the message_info.json

    :param filename: file to load messages from
    :type filename: str
    :return: No explicit return, but loads messages into _LOADED_MESSAGES
    :rtype: None
    """
    global _LOADED_MESSAGES
    if not filename:
        directory, _ = os.path.split(__file__)
        filename = os.path.join(directory, 'message_info.json')
    with open(filename, 'r') as f:
        _LOADED_MESSAGES = json.load(f)


def get_message(primary_number, secondary_number='default'):
    """
    Gets the message for a primary number and a secondary number for a message
    categorization.  If the rule doesn't have the secondary number in
    existence it grabs the default value for the primary categorization.

    :param primary_number: The primary number for the rule.
    :type primary_number: str
    :param: The secondary number for the rule.
    :type secondary_number: str
    :return: The message text, or None if the primary number does not exist
    :rtype: None or str
    """
    global _LOADED_MESSAGES
    if not _LOADED_MESSAGES:
        _load_messages()
    if primary_number not in _LOADED_MESSAGES:
        return '(Message code not found)'
    return _LOADED_MESSAGES[primary_number].get(
        secondary_number, _LOADED_MESSAGES[primary_number]['default']
    )



# (c) Copyright 2015 - 2016, XBRL US Inc. All rights reserved.
# See license.md for license information.
# See PatentNotice.md for patent infringement notice.
from .util import facts, messages

from arelle.ModelValue import dateunionDate


_CODE_NAME_33 = 'DQC.US.0033'
_CODE_NAME_36 = 'DQC.US.0036'
_RULE_VERSION = '1.0'


def doc_period_end_date_check(val):
    """
    Creates an error with all the relevant information from each of the bad
    DocumentPeriodEndDates returned from _doc_period_end_date_check

    :param val: val from which to throw an error
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: No explicit return, Throws error for everything returned by
        _doc_period_end_date_check
    :rtype: None
    """
    for params in _doc_period_end_date_check(val.modelXbrl):
        code, message, context_date, mod_obj, default_dped_fact = params
        val.modelXbrl.error(
            code, message,
            dped_context_date=context_date,
            modelObject=(mod_obj, default_dped_fact),
            ruleVersion=_RULE_VERSION
        )


def _is_valid_eop_fact(eop_fact):
    """
    Checks to ensure that the eop_fact and the eop_fact.xValue is not None
    :param eop_fact: fact to check
    :type eop_fact: :class:'~arelle.InstanceModelObject.ModelFact'
    :return: True if the fact and the fact.xValue is not None
    :rtype: bool
    """
    if eop_fact is None:
        return False
    elif eop_fact.xValue is None:
        return False
    return True


def _doc_period_end_date_check(model_xbrl):
    """
    Compares the value of DocumentPeriodEndDate against the end date of its
    context. If the difference is more than 3 days, fires a validation error.
    For each DocumentPeriodEndDate, if the above check doesn't fire, check all
    DEI fact context end dates against it.

    :param model_xbrl: ModelXbrl to check facts
    :type model_xbrl: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: list of tuples containing bad DocumentPeriodEndDates
    :rtype: list [tuple]
    """
    dped_facts, dei_facts = _setup_dei_facts(model_xbrl)
    default_dped_fact = _get_default_dped(dped_facts)
    result_group = []
    if default_dped_fact is None:
        return result_group

    not_valid_dped = []
    # loop through the DocumentPeriodEndDate's to check for
    # consistent dates
    for eop_facts in dped_facts.values():
        eop_fact = eop_facts[0]
        eop_context = eop_fact.context
        if ((not _is_valid_eop_fact(eop_fact) or
             eop_context is None or
             eop_context.endDatetime is None)):
            continue
        fact_eop_date = dateunionDate(eop_fact.xValue, subtractOneDay=True)
        # Arelle adjusts context end date to end-of-day midnight
        # Reverse the adjustment to get the expected date value
        # by subtracting one day
        context_eop_date = dateunionDate(
            eop_context.endDatetime,
            subtractOneDay=True
        )
        delta = context_eop_date - fact_eop_date
        if abs(delta.days) > 3:
            for axis, dim_value in eop_fact.context.segDimValues.items():
                if axis.qname.localName == 'LegalEntityAxis':
                    not_valid_dped.append(dim_value.memberQname.localName)
            result_group.append((
                '{}.1'.format(_CODE_NAME_36),
                messages.get_message(_CODE_NAME_36),
                context_eop_date,
                eop_fact,
                default_dped_fact
            ))

    # Don't loop through them if the DPED date is bad, since the date
    # is incorrect.

    # loop through the dei facts and compare against their LEA's
    # DocumentPeriodEndDate
    for lea_key, fact_group in dei_facts.items():
        eop_fact = dped_facts.get(lea_key, default_dped_fact)[0]
        if ((eop_fact is None or
             eop_fact.context is None or
             eop_fact.context.endDatetime is None
             )):
            continue

        # Arelle adjusts context end date to end-of-day midnight
        # Reverse the adjustment to get the expected date value
        # by subtracting one day
        context_eop_date = dateunionDate(
            eop_fact.context.endDatetime, subtractOneDay=True
        )

        if len(fact_group) > 0:
            # Check all DEI facts against this DocumentPeriodEndDate.
            # If the DocumentPeriodEndDate context check doesn't fire,
            # we will check all dei fact context end dates against it.
            for fact in fact_group:
                if ((fact.context is None or
                     fact.context.endDatetime is None or
                     fact.concept.periodType != 'duration'
                     )):
                    continue

                if check_for_lea_member(fact, not_valid_dped):
                    delta = context_eop_date - dateunionDate(
                        fact.context.endDatetime, subtractOneDay=True
                    )
                    if delta.days != 0 and abs(delta.days) <= 3:
                        result_group.append((
                            '{}.2'.format(_CODE_NAME_33),
                            messages.get_message(_CODE_NAME_33),
                            fact.concept.label(),
                            fact,
                            default_dped_fact
                        ))
    return result_group


def check_for_lea_member(fact, not_valid_dped):
    """
    Checks facts to determine whether the fact contains a LEA member that we
    do not want to fire rule 33 for

    :param fact: fact to check
    :type fact: :class:'~arelle.InstanceModelObject.ModelFact'
    :param not_valid_dped: list of LEA member's that we do not want to fire
        rule 33 on
    :type not_valid_dped: list
    :return: False (so we don't continue) if the fact contains a LEA member
        that is in the list of LEA members we do not want to fire rule 33 for.
        True (so we do continue) otherwise.
    :rtype: bool
    """
    for fact_axis, fact_dim_value in fact.context.segDimValues.items():
        if fact_dim_value.memberQname.localName in not_valid_dped:
            # If we find the member, we do not want to continue with rule 33
            # check
            return False
    return True


def _setup_dei_facts(model_xbrl):
    """
    Sets up the dei facts using the specified modelXbrl

    :param model_xbrl: ModelXbrl to get the dei facts from
    :type model_xbrl: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: a tuple of the dictionary of a list of 1 DocumentPeriodEndDate per
        LegalEntityAxis and the dictionary of the list of dei facts per
        LegalEntityAxis
    :rtype: tuple (dict, dict)
    """
    ignored_fact_list = [
        'EntityCommonStockSharesOutstanding',
        'EntityPublicFloat',
        'DocumentPeriodEndDate',
        'EntityNumberOfEmployees',
        'EntityListingDepositoryReceiptRatio'
    ]

    dei_facts = facts.legal_entity_axis_facts_by_member(
        _get_dei_facts(model_xbrl, ignored_fact_list)
    )
    dped_facts = facts.legal_entity_axis_facts_by_member(
        facts.get_facts_dei(['DocumentPeriodEndDate'], model_xbrl)
    )

    return dped_facts, dei_facts


def _get_dei_facts(model_xbrl, exclude_list=[None]):
    """
    Returns a list of all the modelXbrl's DEI facts that aren't in the
    exclude_list; this can be fed into prepare_facts_for_calculation.

    :param model_xbrl: ModelXbrl to gather DEI facts from
    :type model_xbrl: :class:'~arelle.ModelXbrl.ModelXbrl'
    :param exclude_list: names of facts to exclude
    :type exclude_list: list [str]
    :return: list of all the modelXbrl DEI facts that aren't in exclude list
    :rtype: list [:class:'~arelle.InstanceModelObject.ModelFact']
    """
    return [
        f for f in model_xbrl.facts
        if f.namespaceURI in facts.DEI_NAMESPACE_LIST and
        f.qname.localName not in exclude_list
    ]


def _get_default_dped(dped_facts):
    """
    Returns the default DocumentPeriodEndDate fact. If there are no
    DocumentPeriodEndDate facts it returns None. If the DocumentPeriodEndDate
    cannot be figured it returns a list of None.

    :param dped_facts: dictionary of facts to get DocumentEndPeriodDate from
    :type dped_facts: dict
    :return: the default DocumentPeriodEndDate
    :rtype: str
    """
    keys = dped_facts.keys()
    if len(keys) == 0:
        return None
    elif len(keys) == 1:
        # Only one Document Period End Date, so that is our default
        value, = dped_facts.values()
        return value
    else:
        return dped_facts.get('', [None])

__pluginInfo__ = {
    'name': '{}, {}'.format(_CODE_NAME_33, _CODE_NAME_36),
    'version': _RULE_VERSION,
    'description': 'Checks the doc period end date relative to fact contexts.',
    # Mount points
    'Validate.XBRL.Finally': doc_period_end_date_check,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages


readme = open('README.md').read()


def get_version():
    import imp

    source_dir = 'dqc_us_rules'

    with open('{}/_pkg_meta.py'.format(source_dir), 'rb') as fp:
        mod = imp.load_source('_pkg_meta', source_dir, fp)

    return mod.version


setup_args = dict(
    name='dqc_us_rules',
    version=get_version(),
    description="""""".strip(),
    long_description=readme,
    maintainer='Data Quality Committee',
    maintainer_email='dqc@xbrl.us',
    url='https://github.com/DataQualityCommittee/dqc_us_rules',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: Other/Proprietary License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.4',
        "Topic :: Office/Business :: Financial :: Accounting"
    ],
    test_suite='test',
    install_requires=[],
    tests_require=[
        "lxml==2.3.5",
        "nose==1.3.4",
        "mock==1.0.1",
        "arelle"
    ],
)


if __name__ == '__main__':
    setup(**setup_args)



# (c) Copyright 2015 - 2016, XBRL US Inc. All rights reserved.
# See license.md for license information.
# See PatentNotice.md for patent infringement notice.
from collections import defaultdict
from datetime import timedelta
from .util import facts, messages
import csv
import os

CHECK_TYPES = ['textBlockItemType']
CHECK_DEI = [
    'AmendmentDescription', 'AmendmentFlag', 'CurrentFiscalYearEndDate',
    'DocumentPeriodEndDate', 'DocumentFiscalYearFocus',
    'DocumentFiscalPeriodFocus', 'DocumentType', 'EntityRegistrantName',
    'EntityCentralIndexKey', 'EntityFilerCategory'
]
_CODE_NAME = 'DQC.US.0006'
_RULE_VERSION = '1.0'
_DEFAULT_DATE_BOUNDS_FILE = os.path.join(
    os.path.dirname(__file__),
    'resources',
    'DQC_US_0006',
    'dqc_06_date_bounds.csv'
)


def validate_dates_within_periods(val):
    """
    Checks date ranges are within expected values for the fiscal focus period

    :param val: val to check
    :type val: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: No direct return, but throws errors when dates aren't validated
    :rtype: none
    """
    date_bounds_dict = _date_bounds_from_csv()
    doc_type = facts.lookup_dei_facts('DocumentType', val.modelXbrl)
    if len(doc_type) != 1 or 'T' in doc_type[0].xValue:
        # If it is a transitional document, or there is more than one
        # document type declared, we will not run this check.
        return
    dict_of_facts = (
        _date_range_check(
            CHECK_TYPES, CHECK_DEI, date_bounds_dict, val.modelXbrl
        )
    )
    for document_fiscal_period_focus, fact_list in dict_of_facts.items():
        for fact in fact_list:
            val.modelXbrl.error(
                '{}.14'.format(_CODE_NAME),
                messages.get_message(_CODE_NAME),
                concept=fact.qname,
                period=document_fiscal_period_focus.xValue,
                modelObject=[fact, document_fiscal_period_focus],
                ruleVersion=_RULE_VERSION
            )


def _date_range_check(check_types, check_dei, date_bounds_dict, model_xbrl):
    """
    Takes two lists of fact names, a dict of date boundaries and modelXbrl and
    then compiles a list of all facts in the modelXbrl that match the names in
    the supplied name lists. It then compares the context date span to the date
    boundaries for the corresponding document period focus. Any facts with
    spans less than or larger than the supplied boundaries are returned in a
    dict based on the document period focus.

    :param check_types: first list of names
    :type check_types: list [str]
    :param check_dei: second list of names
    :type check_dei: list [str]
    :param date_bounds_dict: period from which names should be pulled from
    :type date_bounds_dict: dict
    :param model_xbrl: model xbrl from which facts are to be pulled from
    :type model_xbrl: :class:'~arelle.ModelXbrl.ModelXbrl'
    :return: list of facts that match the names in the two lists and are within
        the date_bounds_dict
    :rtype: dict
    """
    facts_in_error = defaultdict(list)
    list_of_facts = facts.legal_entity_axis_facts_by_member(
        facts.get_facts_with_type(check_types, model_xbrl)
    )
    list_of_facts = _dict_list_update(
        list_of_facts,
        facts.legal_entity_axis_facts_by_member(
            facts.get_facts_dei(check_dei, model_xbrl)
        )
    )

    dfpf_list = facts.lookup_dei_facts('DocumentFiscalPeriodFocus', model_xbrl)
    dfpf_dict = facts.legal_entity_axis_facts_by_member(dfpf_list)
    for lea_member, fact_list in list_of_facts.items():
        lookup = (
            lea_member
            if lea_member in dfpf_dict
            else facts.LEGALENTITYAXIS_DEFAULT
        )
        if lookup in dfpf_dict:
            focus_l = set([
                foc for foc in dfpf_dict[lookup]
                if foc.xValue in date_bounds_dict
            ])

            if len(focus_l) != 1:
                continue
            focus = focus_l.pop()
            min_span = (
                timedelta(days=date_bounds_dict[focus.xValue].get('min'))
            )
            max_span = (
                timedelta(days=date_bounds_dict[focus.xValue].get('max'))
            )

            for fact in fact_list:
                if ((fact.context.endDatetime is not None and
                     fact.context.startDatetime is not None)):
                    span = (
                        fact.context.endDatetime - fact.context.startDatetime
                    )
                    if span < min_span or span > max_span:
                        facts_in_error[focus].append(fact)
    return facts_in_error


def _dict_list_update(dict_a, dict_b):
    """
    Helper for the LEA dictionaries, extends the lists from dict_a with the
    lists in dict_b.

    :param dict_a: dictionary to be extended
    :type dict_a: dict
    :param dict_b: dictionary to extend into other dictionary
    :type dict_b: dict
    :return: lists in dict_a extended with lists in dict_b
    :rtype: dict
    """
    for key, val in dict_b.items():
        dict_a[key].extend(val)
    return dict_a


def _date_bounds_from_csv():
    """
    Returns a map of {time_period: {'min':min_value,'max':max_value}}
    ex: date_bounds_from_csv()['Q1'] = {'min':65,'max':115}

    :return: A map of {time_period: {'min':min_value,'max':max_value}}.
    :rtype: dict
    """
    with open(_DEFAULT_DATE_BOUNDS_FILE, 'r') as f:
        reader = csv.reader(f)
        date_bounds_dict = {}
        next(reader, None)
        for row in reader:
            date_bounds_dict[row[0]] = {'min': int(row[1]), 'max': int(row[2])}
        return date_bounds_dict

__pluginInfo__ = {
    'name': _CODE_NAME,
    'version': _RULE_VERSION,
    'description': (
        'Checks all of the specified types and concepts for their date '
        'ranges to verify the ranges are within expected paramters for the '
        'fiscal periods'
    ),
    # Mount points
    'Validate.XBRL.Finally': validate_dates_within_periods,
}
